from .b2_visual import *
from .metrics import *
from .folds import *
from .math_funcs import *
from .plot_classification import *
from .file_handling import *


AF = B2 + METRICS + FOLDS + MATH_FUNCS + PLOT_CLASSIFICATION + FH


from ...forall import *
#######################################################################################################################
# Нормальные случайные векторы
#######################################################################################################################
def NRV_1(muX,muY,sigmaX2,sigmaY2,rho,Px,Py):
    """Для нормального случайного вектора<br>
    (X,Y) ∼ N(muX, muY, sigmaX2, sigmaY2, rho)<br>
    Найдите вероятность P((X - Px)(Y - Py) < 0).

    Args:
        muX (numerical): Математическое ожидание X
        muY (numerical): Математическое ожидание Y
        sigmaX2 (numerical): Дисперсия X
        sigmaY2 (numerical): Дисперсия Y
        rho (numerical): Коэффициент корреляции между X и Y
        Px (numerical): Число для вычитания из X
        Py (numerical): Число для вычитания из Y

    ## Prints
        P((X - Px)(Y - Py) < 0) = `answer` с запятой вместо точки и сокращением до 4 знаков после запятой
    
    Returns:
        answer (numerical): Вероятность P((X - Px)(Y - Py) < 0)
    """
    ''''''
    
    import math
    import numpy as np
    import scipy.stats
    
    N = {'muX': muX, 'muY': muY, 'sigmaX2': sigmaX2, 'sigmaY2': sigmaY2, 'rho': rho}
    mu = np.array([N['muX'], N['muY']])
    Cov = np.array([[N['sigmaX2'], N['rho']*math.sqrt(N['sigmaX2'])*math.sqrt(N['sigmaY2'])],
                    [N['rho']*math.sqrt(N['sigmaX2'])*math.sqrt(N['sigmaY2']), N['sigmaY2']]])
    W = scipy.stats.multivariate_normal(mu, Cov)
    X = scipy.stats.norm(N['muX'], math.sqrt(N['sigmaX2']))
    Y = scipy.stats.norm(N['muY'], math.sqrt(N['sigmaY2']))
    Pa=X.cdf(Px)-W.cdf([Px,Py])
    Pb=Y.cdf(Py)-W.cdf([Px,Py])
    
    answer=Pa+Pb
    
    print(f'P((X-{Px})(Y-{Py})<0) = {rrstr(answer,4)}')
    
    return answer
#######################################################################################################################
def NRV_2(muX,muY,sigmaX2,sigmaY2,rho,xminus1,xminus2,yminus):
    """Для нормального случайного вектора <br>
    (X,Y) ∼ N(muX,muY,sigmaX2,sigmaY2,rho)<br>
    Найдите вероятность P((X - xminus1)(X - xminus2)(Y - yminus) < 0).

    Args:
        muX (numerical): Математическое ожидание X
        muY (numerical): Математическое ожидание Y
        sigmaX2 (numerical): Дисперсия X
        sigmaY2 (numerical): Дисперсия Y
        rho (numerical): Коэффициент корреляции между X и Y
        xminus1 (numerical): Первое число для вычитания из X
        xminus2 (numerical): Второе число для вычитания из X
        yminus (numerical): Число для вычитания из Y

    ## Prints
        P((X - xminus1)(X - xminus2)(Y - yminus) < 0) = `answer` с запятой вместо точки и сокращением до 4 знаков после запятой

    Returns:
        answer (numerical): Вероятность P((X - xminus1)(X - xminus2)(Y - yminus) < 0)
    """ 
    import numpy as np
    import scipy.stats
    
    np.abs
    
    sigmaX = sigmaX2**0.5
    sigmaY = sigmaY2**0.5
    
    mu = np.array([muX,muY])
    Cov = np.array([[sigmaX**2, rho*sigmaX*sigmaY], [rho*sigmaX*sigmaY, sigmaY**2]])
    
    X = scipy.stats.norm(muX, sigmaX)
    Y = scipy.stats.norm(muY, sigmaY)
    W = scipy.stats.multivariate_normal(mu, Cov)
    
    Pa = W.cdf([xminus1, yminus])
    Pb = X.cdf(xminus2) - X.cdf(xminus1) - (W.cdf([xminus2, yminus]) - W.cdf([xminus1, yminus]))
    Pc = Y.cdf(yminus) - W.cdf([xminus2, yminus])
    
    answer = Pa+Pb+Pc
    
    print(f'P((X - {xminus1})(X - {xminus2})(Y - {yminus}) < 0) = {rrstr(answer,4)}')
    
    return answer
#######################################################################################################################
def NRV_3(c_x2 = 0,c_x = 0,c_xy = 0,c_y = 0,c_y2 = 0):
    """Случайный вектор (X,Y) имеет плотность распределения<br>
    f_X,Y(x,y) = ( [невлияющий коэф.] * e^-1/2*(c_x2 * x^2 + c_x * x + c_xy * xy + c_y * y + c_y2 * y^2 + [невлияющая константа]) / π
    
    Найдите:
    - математическое ожидание E(X)
    - математическое ожидание E(Y)
    - дисперсию  Var(X)
    - дисперсию  Var(Y)
    - ковариацию  Cov(X,Y)
    - коэффициент корреляции  ρ(X,Y)

    Args:
        c_x2 (numerical, optional): Коэффициент при x^2. По умолчанию 0.
        c_x (numerical, optional): Коэффициент при x. По умолчанию 0.
        c_xy (numerical, optional): Коэффициент при xy. По умолчанию 0.
        c_y (numerical, optional): Коэффициент при y. По умолчанию 0.
        c_y2 (numerical, optional): Коэффициент при y^2. По умолчанию 0.

    ## Prints
        EX = `sol[EX]`, EY = `sol[EY]`, VarX = `VarX`, VarY = `VarY`, CovXY = `CovXY`, roXY = `roXY`

    Returns:
        answer (tuple): (`sol[EX]`, `sol[EY]`, `VarX`, `VarY`, `CovXY`, `roXY`)
    """
    import sympy
    #после выноса -1/2!!!!!!
    coefs = {
        'x^2': c_x2,
        'x': c_x,
        'xy': c_xy,
        'y': c_y,
        'y^2': c_y2,
    }
    C = sympy.Matrix([[coefs['x^2'], int(coefs['xy']/2)], [int(coefs['xy']/2), coefs['y^2']]])
    C1 = C**(-1)
    VarX = C1[0, 0]
    sigmaX = sympy.sqrt(VarX)
    VarY = C1[1, 1]
    sigmaY = sympy.sqrt(VarY)
    CovXY = C1[0, 1]
    roXY = CovXY/(sigmaX*sigmaY)
    EX, EY = sympy.symbols('EX, EY')
    equations = (
        sympy.Eq(int(coefs['x^2'])*EX + int(coefs['xy']/2)*EY, int(coefs['x']*(-1/2))),
        sympy.Eq(int(coefs['xy']/2)*EX + int(coefs['y^2'])*EY, int(coefs['y']*(-1/2)))
    )
    sol = sympy.solve(equations, (EX, EY))
    
    print(f'EX = {sol[EX]}, EY = {sol[EY]}, VarX = {VarX}, VarY = {VarY}, CovXY = {CovXY}, roXY = {roXY}')
    
    answer = (sol[EX],sol[EY],VarX,VarY,CovXY,roXY)
    
    return answer
#######################################################################################################################
NRV=[NRV_1,NRV_2,NRV_3] # Список функций в файле


import re
from ..forall import *
from .base import BASE
from .eqs import EQS
from .sistems import SISTEMS
from .interp import INTER
from .eigen import EIGEN
from .strassen import STRASSEN
from .matrices import MATRICES
from .matrices.householder import HOUSEHOLDER
from .matrices.SVD import SVDF
from .differential_eqs import DF



from .additional_funcs import AF


MATRICES = MATRICES + HOUSEHOLDER + SVDF

def printcolab():
    """Выводит ссылку на знакомый гугл колаб(google colab)"""
    print(r'https://colab.research.google.com/drive/1QEu-jq5sgb53x76maYmtV3MvhqeB--zu?usp=sharing')
    return r'https://colab.research.google.com/drive/1QEu-jq5sgb53x76maYmtV3MvhqeB--zu?usp=sharing'



UF = [printcolab] + AF

files_dict ={
    'База':BASE,
    'Решение уравнений': EQS,
    'Системы уравнений' : SISTEMS,
    'Интерполяция' : INTER,
    'Перемножение Штрассена' : STRASSEN,
    'Поиск собственных' : EIGEN,
    'Дифференциальные ур-я': DF,
    
    
    #############################
    'Дополнительные функции': UF,
    'Матричные операции' : MATRICES
    
}

names = list(files_dict.keys())
modules = list(files_dict.values())


funcs_dicts = [dict([(get_task_from_func(i), i) for i in module]) for module in modules]
funcs_dicts_ts = [dict([(get_task_from_func(i,True), i) for i in module]) for module in modules]
funcs_dicts_full = [dict([(i.__name__, getsource(i)) for i in module]) for module in modules]


themes_list_funcs = dict([(names[i],list(funcs_dicts[i].values()) ) for i in range(len(names))]) # Название темы : список функций по теме
themes_list_dicts = dict([(names[i],funcs_dicts[i]) for i in range(len(names))])                 # Название темы : словарь по теме, где ЗАДАНИЕ: ФУНКЦИИ
themes_list_dicts_full = dict([(names[i],funcs_dicts_full[i]) for i in range(len(names))])       # Название темы : словарь по теме, где НАЗВАНИЕ ФУНКЦИИ: ТЕКСТ ФУНКЦИИ



def description(
    dict_to_show=themes_list_funcs,
    key=None,
    show_only_keys: bool = False,
    show_keys_second_level: bool = True,
    n_symbols: int = 32,
    to_print: bool = True):
    """
    Форматированный вывод информации о функциях и заданиях из словарей тем.

    Parameters
    ----------
    dict_to_show : str or dict, optional
        Имя словаря тем или сам словарь вида {функция: задание}.
        По умолчанию используется `themes_list_funcs`.
    key : hashable, optional
        Ключ для фильтрации конкретного элемента словаря.
    show_only_keys : bool, optional
        Если True - показывать только ключи словаря.
    show_keys_second_level : bool, optional
        Если True - показывать ключи второго уровня (названия функций).
    n_symbols : int, optional
        Максимальное количество символов описания для отображения.
    to_print : bool, optional
        Если True - выводить результат через print(), иначе вернуть как строку.

    Returns
    -------
    str or None
        Форматированная строка с информацией (если to_print=False).
        При to_print=True возвращает None, выводя результат напрямую.

    Notes
    -----
    1. Функция поддерживает два режима работы:
       - При передаче строки в dict_to_show: использует предопределенные словари тем
       - При передаче словаря напрямую: работает с пользовательскими структурами
    2. Использует вспомогательную функцию invert_dict() для преобразования словарей.
    3. При обработке описаний автоматически добавляется перенос строк для длинных текстов.
    4. В случае ошибок при извлечении описаний (например, отсутствующий ключ) - пропускает проблемные элементы.

    Examples
    --------
    >>> description('math_operations', show_only_keys=True)
    Сложение  : 
    Вычитание : 
    Умножение : 

    >>> description('data_processing', n_symbols=50)
    clean_data   : Очистка данных от пропусков...
    analyze_data : Проведение статистического анализа...

    >>> description('api_calls', key='get_request')
    get_request : Выполняет GET-запрос к API с параметрами...

    References
    ----------
    .. [1] Python Software Foundation. "Python Language Reference", version 3.11.
    .. [2] Beazley, D.M. "Python Essential Reference", 4th edition.
    .. [3] Ramalho, L. "Fluent Python: Clear, Concise, and Effective Programming".
    """

    if type(dict_to_show) == str and key == None:
        dict_to_show = themes_list_dicts[dict_to_show]
        dict_to_show = invert_dict(dict_to_show)
        text = ""
        length1 = 1 + max([len(x.__name__) for x in list(dict_to_show.keys())])
        
        for key in dict_to_show.keys():
            text += f'{key.__name__:<{length1}}'
            
            if not show_only_keys:
                text += ': '
                text += f'{dict_to_show[key]};\n' + ' '*(length1+2)
            text += '\n'
            
        if to_print == True:
            return print(text)
        return text

    elif type(dict_to_show) == str and key in themes_list_dicts_full[dict_to_show].keys():
        return print(themes_list_dicts_full[dict_to_show][key])
    
    else:
        show_only_keys = False
    text = ""
    length1 = 1 + max([len(x) for x in list(dict_to_show.keys())])
    
    for key in dict_to_show.keys():
        text += f'{key:^{length1}}'
        if not show_only_keys:
            text += ': '
            for f in dict_to_show[key]:
                text += f'{f.__name__}'
                if show_keys_second_level:
                    text += ': '
                    try:
                        func_text_len = len(invert_dict(themes_list_dicts[key])[f])
                        func_text = invert_dict(themes_list_dicts[key])[f]
                        text += func_text.replace('\n','\n'+' '*(length1 + len(f.__name__))) if func_text_len < n_symbols else func_text[:n_symbols].replace('\n','\n'+' '*(length1 + len(f.__name__)))+'...'
                    except:
                        pass
                text += ';\n' + ' '*(length1+2) # + '\n' + ' '*(length1+2)
        text += '\n'
        
    if to_print == True:
        return print(text)
    return text



def search(query: str, to_print: bool = True, data: str = description(n_symbols=10000, to_print=False)):
    """
    Выполняет поиск совпадений по запросу в структурированных данных тем и функций.

    Parameters
    ----------
    query : str
        Строка для поиска. Искать можно по части слова или фразы.
    to_print : bool, optional
        Если True - результаты выводятся через print(). По умолчанию True.
    data : str, optional
        Исходные данные для поиска в формате, возвращаемом функцией description().
        По умолчанию вызывает description(n_symbols=10000, to_print=False).

    Returns
    -------
    list or None
        При to_print=True: выводит результаты построчно через print()
        При to_print=False: возвращает список строк в формате "Тема : Описание"

    Notes
    -----
    1. Функция чувствительна к структуре входных данных:
       - Темы должны быть разделены двойными переносами строк
       - Каждая тема должна начинаться с заголовка в формате "Название_темы:"
       - Функции внутри темы должны быть в формате "имя_функции: описание"
    2. Поиск регистронезависимый
    3. Рекомендуется использовать n_symbols=10000 при вызове description() 
       для сохранения полноты описаний

    Examples
    --------
    >>> search("обработка данных")
    math_operations : Очистка данных от пропусков...
    data_processing : Обработка данных с помощью pandas...

    >>> search("ML", to_print=False)
    ['ai_tools : Машинное обучение с использованием sklearn', 
     'deep_learning : Реализация нейронных сетей']

    >>> search("API", data="custom_data_string")
    api_calls : Выполнение GET-запросов к REST API

    References
    ----------
    .. [1] Внутренняя документация проекта, описывающая структуру тем и функций
    .. [2] Python Software Foundation. "Python Language Reference", version 3.11.
    """
    topics = re.split(r'\n\s*\n', data)
    matches = []

    for topic_data in topics:
        if not topic_data.strip():
            continue

        topic_match = re.match(r'^\s*(.*?):', topic_data)
        if not topic_match:
            continue

        topic = topic_match.group(1).strip()
        functions = re.findall(r'(\w+)\s*:\s*([\s\S]*?)(?=\n\s*\w+\s*:|\Z)', topic_data)

        for func, description in functions:
            if query.lower() in description.lower():
                matches.append(f"{topic} : {description.strip()}")

    if to_print:
        return print("\n".join(matches))
    return matches


from ....forall import *
from ...additional_funcs import parallel_max_offdiag
####################################################################################
def jacobi_eigen(A, tol=1e-10, max_iter=20000):
    """
    Вычисление собственных значений и собственных векторов симметричной матрицы методом Якоби с векторизованными операциями.
    Вовсе не самая быстрая из реализаций этого алгоритма 
    
    Теоретическая часть:
    Метод Якоби основан на последовательных вращениях матрицы для обнуления внедиагональных элементов. 
    Каждое вращение определяется углом, который минимизирует текущий внедиагональный элемент. 
    При достаточном количестве итераций матрица преобразуется к диагональному виду, где диагональные элементы становятся собственными значениями, а столбцы матрицы поворотов — собственными векторами [[3]][[4]][[7]].

    Практическая реализация:
    - Предварительное кэширование индексов верхней треугольной матрицы для ускорения поиска максимального внедиагонального элемента.
    - Векторизованные операции обновления матрицы для повышения производительности.
    - Обновление матрицы собственных векторов через последовательные повороты.
    
    Parameters
    ----------
    A : np.ndarray, shape (n, n)
        Входная симметричная матрица, для которой необходимо найти собственные значения и векторы.
        Предполагается, что матрица симметричная. Несимметричные матрицы могут привести к некорректным результатам.
    tol : float, optional (default=1e-10)
        Порог точности для завершения итераций. Итерации останавливаются, если максимальный внедиагональный элемент меньше `tol`.
    max_iter : int, optional (default=20000)
        Максимальное количество итераций алгоритма.

    Returns
    -------
    eigenvalues : np.ndarray, shape (n,)
        Массив собственных значений матрицы `A`, отсортированных по убыванию.
    V : np.ndarray, shape (n, n)
        Матрица, столбцы которой являются собственными векторами, соответствующими собственным значениям.

    Examples
    --------
    >>> import numpy as np
    >>> A = np.array([[4, 1], [1, 3]], dtype=np.float64)
    >>> eigenvalues, eigenvectors = jacobi_eigen(A)
    >>> print("Собственные значения:", eigenvalues)
    >>> print("Собственные векторы:\n", eigenvectors)

    Notes
    -----
    1. Функция предполагает, что матрица `A` симметричная. Несимметричные матрицы могут привести к некорректным результатам.
    2. Используется классический метод Якоби с последовательным обнулением внедиагональных элементов через вращения Гивенса.
    3. Поиск максимального внедиагонального элемента выполняется с использованием `np.triu_indices` и `np.argmax`, что ускоряет процесс по сравнению с вложенными циклами [[4]].
    4. Обновление матриц `A` и `V` реализовано с использованием векторизованных операций NumPy для повышения производительности [[5]].
    5. Алгоритм имеет сложность O(n³), но за счет векторизации демонстрирует хорошую скорость на малых и средних матрицах.

    References
    ----------
    .. [4] "5 minute guide to Numba" - https://numba.readthedocs.io/
    .. [5] Golub, G. H., & Van Loan, C. F. (2013). Matrix computations. Johns Hopkins University Press.
    """
    A = np.array(A, dtype=np.float64)
    n = A.shape[0]
    V = np.eye(n)
    
    # Предварительное кэширование индексов
    indices = np.triu_indices(n, k=1)
    
    for _ in (range(max_iter)):
        # Ускоренный поиск максимального внедиагонального элемента
        triu_abs = np.abs(A[indices])
        max_idx = np.argmax(triu_abs)
        max_val = triu_abs[max_idx]
        
        if max_val < tol:
            break
            
        i, j = indices[0][max_idx], indices[1][max_idx]
        
        # Вычисление угла поворота
        alpha = (A[j,j] - A[i,i]) / (2 * A[i,j])
        t = np.sign(alpha) / (np.abs(alpha) + np.sqrt(1 + alpha**2))
        c = 1 / np.sqrt(1 + t**2)
        s = t * c
        
        # Векторизованное обновление матрицы A
        row_i = A[i, :].copy()
        row_j = A[j, :].copy()
        
        A[i, :] = c*row_i - s*row_j
        A[j, :] = s*row_i + c*row_j
        A[:, i] = A[i, :]
        A[:, j] = A[j, :]
        
        A[i,j] = 0
        A[j,i] = 0
        
        # Обновление собственных векторов
        V_i = V[:, i].copy()
        V_j = V[:, j].copy()
        V[:, i] = c*V_i - s*V_j
        V[:, j] = s*V_i + c*V_j
        
    eigenvalues = np.diag(A)
    return eigenvalues, V
####################################################################################
@njit
def numba_jacobi_eigen(A, tol=1e-10, max_iter=20000):
    """
    Вычисление собственных значений и собственных векторов симметричной матрицы методом Якоби с ускорением через Numba.
    Несамая быстрая из реализаций этого алгоритма 
    Теоретическая часть:
    Метод Якоби основан на последовательных вращениях матрицы для обнуления внедиагональных элементов. 
    Каждое вращение определяется углом, который минимизирует текущий внедиагональный элемент. 
    При достаточном количестве итераций матрица преобразуется к диагональному виду, где диагональные элементы становятся собственными значениями, а столбцы матрицы поворотов — собственными векторами [[3]][[4]][[7]].

    Практическая реализация:
    - Предварительное кэширование индексов верхней треугольной матрицы для ускорения поиска максимального внедиагонального элемента.
    - Векторизованные операции обновления матрицы для повышения производительности.
    - Обновление матрицы собственных векторов через последовательные повороты.
    
    Parameters
    ----------
    A : np.ndarray, shape (n, n)
        Входная симметричная матрица, для которой необходимо найти собственные значения и векторы.
        Предполагается, что матрица симметричная. Несимметричные матрицы могут привести к некорректным результатам.
    tol : float, optional (default=1e-10)
        Порог точности для завершения итераций. Итерации останавливаются, если максимальный внедиагональный элемент меньше `tol`.
    max_iter : int, optional (default=20000)
        Максимальное количество итераций алгоритма.

    Returns
    -------
    eigenvalues : np.ndarray, shape (n,)
        Массив собственных значений матрицы `A`, отсортированных по убыванию.
    V : np.ndarray, shape (n, n)
        Матрица, столбцы которой являются собственными векторами, соответствующими собственным значениям.

    Examples
    --------
    >>> import numpy as np
    >>> A = np.array([[4, 1], [1, 3]], dtype=np.float64)
    >>> eigenvalues, eigenvectors = numba_jacobi_eigen(A)
    >>> print("Собственные значения:", eigenvalues)
    >>> print("Собственные векторы:\n", eigenvectors)

    Notes
    -----
    1. Функция предполагает, что матрица `A` симметричная. Несимметричные матрицы могут привести к некорректным результатам.
    2. Используется классический метод Якоби с последовательным обнулением внедиагональных элементов через вращения Гивенса.
    3. Поиск максимального внедиагонального элемента выполняется без использования `np.triu_indices` для совместимости с Numba [[1]].
    4. Для повышения производительности используется `@njit` для JIT-компиляции Python-кода в машинный код [[5]].
    5. Алгоритм имеет сложность O(n³), но за счет эффективной реализации на малых матрицах демонстрирует высокую скорость.

    References
    ----------
    .. [1] Numba documentation - https://numba.readthedocs.io/
    .. [2] Golub, G. H., & Van Loan, C. F. (2013). Matrix computations. Johns Hopkins University Press.
    .. [5] "Making Python extremely fast with Numba: Advanced Deep Dive (1/3)" - https://towardsdatascience.com/making-python-extremely-fast-with-numba-advanced-deep-dive-1-of-3-8cb13ed4f5b9
    """
    # Гарантируем, что входная матрица имеет тип float64
    A = A.astype(np.float64)  # Замена np.array на .astype() [[1]]
    n = A.shape[0]
    V = np.eye(n)
    
    for _ in range(max_iter):
        # Поиск максимального внедиагонального элемента без np.triu_indices
        max_val = 0.0
        i_max, j_max = 0, 0
        for i in range(n):
            for j in range(i+1, n):
                current = abs(A[i, j])
                if current > max_val:
                    max_val = current
                    i_max, j_max = i, j
        
        if max_val < tol:
            break
        
        i, j = i_max, j_max
        
        # Вычисление угла поворота
        alpha = (A[j,j] - A[i,i]) / (2 * A[i,j] + 1e-16)  # Защита от деления на ноль
        t = np.sign(alpha) / (np.abs(alpha) + np.sqrt(1 + alpha**2))
        c = 1 / np.sqrt(1 + t**2)
        s = t * c
        
        # Обновление матрицы A
        A_i = A[i, :].copy()
        A_j = A[j, :].copy()
        
        A[i, :] = c*A_i - s*A_j
        A[j, :] = s*A_i + c*A_j
        A[:, i] = A[i, :]
        A[:, j] = A[j, :]
        
        A[i, j] = 0.0
        A[j, i] = 0.0
        
        # Обновление собственных векторов
        V_i = V[:, i].copy()
        V_j = V[:, j].copy()
        V[:, i] = c*V_i - s*V_j
        V[:, j] = s*V_i + c*V_j
        
    eigenvalues = np.diag(A)
    return eigenvalues, V
####################################################################################
@njit(parallel=True, cache=True, fastmath=True)
def numba_jacobi_eigen_parallel(A, tol=1e-10, max_iter=20000):
    """
    Вычисление собственных значений и собственных векторов симметричной матрицы методом Якоби с параллельными вычислениями.
    Самая быстрая из реализаций этого алгоритма 
    Теоретическая часть:
    Метод Якоби основан на последовательных вращениях матрицы для обнуления внедиагональных элементов. 
    Каждое вращение определяется углом, который минимизирует текущий внедиагональный элемент. 
    При достаточном количестве итераций матрица преобразуется к диагональному виду, где диагональные элементы становятся собственными значениями, а столбцы матрицы поворотов — собственными векторами [[3]][[4]][[7]].

    Практическая реализация:
    - Предварительное кэширование индексов верхней треугольной матрицы для ускорения поиска максимального внедиагонального элемента.
    - Векторизованные операции обновления матрицы для повышения производительности.
    - Обновление матрицы собственных векторов через последовательные повороты.

    Parameters
    ----------
    A : np.ndarray, shape (n, n)
        Входная симметричная матрица, для которой необходимо найти собственные значения и векторы.
    tol : float, optional (default=1e-10)
        Порог точности для завершения итераций. Итерации останавливаются, если максимальный внедиагональный элемент меньше `tol`.
    max_iter : int, optional (default=20000)
        Максимальное количество итераций алгоритма.

    Returns
    -------
    eigenvalues : np.ndarray, shape (n,)
        Массив собственных значений матрицы `A`.
    V : np.ndarray, shape (n, n)
        Матрица, столбцы которой являются собственными векторами, соответствующими собственным значениям.

    Examples
    --------
    >>> import numpy as np
    >>> A = np.array([[4, 1], [1, 3]], dtype=np.float64)
    >>> eigenvalues, eigenvectors = numba_jacobi_eigen_parallel(A)
    >>> print("Собственные значения:", eigenvalues)
    >>> print("Собственные векторы:\n", eigenvectors)

    Notes
    -----
    1. Функция предполагает, что матрица `A` симметричная. Несимметричные матрицы могут привести к некорректным результатам.
    2. Используется метод Якоби с последовательным обнулением внедиагональных элементов через вращения Гивенса.
    3. Параллельные вычисления реализованы через `prange` (Numba) для ускорения обработки больших матриц [[5]].
    4. Для повышения производительности включены опции `fastmath=True` и `cache=True`, что может повлиять на численную точность.
    5. Алгоритм имеет сложность O(n³), но за счет параллелизма снижает время выполнения на многоядерных процессорах [[1]].

    References
    ----------
    .. [1] Numba documentation - https://numba.readthedocs.io/
    .. [2] Golub, G. H., & Van Loan, C. F. (2013). Matrix computations. Johns Hopkins University Press.
    .. [3] Parallel computing with Numba - https://numba.pydata.org/numba-doc/latest/user/parallel.html
    """
    # Приводим тип матрицы к float64 для совместимости с numba
    A = A.astype(np.float64)
    n = A.shape[0]
    V = np.eye(n, dtype=np.float64)
    
    for it in range(max_iter):
        # Параллельный поиск максимального вне диагонального элемента
        max_val, i, j = parallel_max_offdiag(A, n)
        if max_val < tol:
            break
        
        # Вычисление угла поворота
        aii = A[i, i]
        ajj = A[j, j]
        aij = A[i, j]
        alpha = (ajj - aii) / (2 * aij + 1e-16)  # защита от деления на 0
        t = np.sign(alpha) / (np.abs(alpha) + np.sqrt(1 + alpha * alpha))
        c = 1 / np.sqrt(1 + t * t)
        s = t * c
        
        # Обновление элементов матрицы A (кроме i и j)
        for k in prange(n):
            if k != i and k != j:
                aik = A[i, k]
                ajk = A[j, k]
                A[i, k] = c * aik - s * ajk
                A[j, k] = s * aik + c * ajk
                A[k, i] = A[i, k]
                A[k, j] = A[j, k]
        
        # Обновление диагональных элементов и зануление вращаемых элементов
        A[i, i] = c * c * aii - 2 * s * c * aij + s * s * ajj
        A[j, j] = s * s * aii + 2 * s * c * aij + c * c * ajj
        A[i, j] = 0.0
        A[j, i] = 0.0
        
        # Обновление матрицы собственных векторов V
        for k in prange(n):
            vik = V[k, i]
            vjk = V[k, j]
            V[k, i] = c * vik - s * vjk
            V[k, j] = s * vik + c * vjk
    
    # Извлечение диагональных элементов (собственных значений)
    eigenvalues = np.empty(n, dtype=A.dtype)
    for i in range(n):
        eigenvalues[i] = A[i, i]
    
    return eigenvalues, V
####################################################################################
JACOBI = [jacobi_eigen, numba_jacobi_eigen, numba_jacobi_eigen_parallel]


import numpy as np
import matplotlib.pyplot as plt
from matplotlib import gridspec
from scipy.stats import gaussian_kde

def plot_classification_2d(
    X, y, model,
    feature_indices=(0, 1),
    feature_names=None,
    class_names=None,
    cmap='viridis',
    scatter_kwargs=None,
    legend_kwargs=None
):
    """
    Визуализация классификации в 2D пространстве с плотностями
    
    Параметры:
        X (array-like): Матрица признаков
        y (array-like): Вектор меток классов
        model: Обученная модель с методом predict_proba или decision_function
        feature_indices (tuple): Индексы двух признаков для визуализации
        feature_names (list): Названия признаков (для подписей)
        class_names (list): Названия классов (для легенды)
        cmap (str): Цветовая карта для контуров
        scatter_kwargs (dict): Параметры для точек (размер, прозрачность)
        legend_kwargs (dict): Параметры легенды
    """
    # Проверка корректности feature_names
    if feature_names is not None:
        if len(feature_names) != 2:
            raise ValueError("feature_names должен содержать ровно 2 элемента")
    
    # Преобразование данных
    X = np.array(X)
    y = np.array(y).ravel()
    if X.ndim == 1:
        X = X.reshape(-1, 1)
    if X.shape[1] < 2:
        raise ValueError("Требуется как минимум 2 признака")

    # Проверка индексов признаков
    idx1, idx2 = feature_indices
    if idx1 >= X.shape[1] or idx2 >= X.shape[1]:
        raise ValueError("Неверные индексы признаков")

    # Подготовка данных
    X_subset = X[:, [idx1, idx2]]
    classes = np.unique(y)
    n_classes = len(classes)
    
    # Генерация сетки
    x_min, x_max = X_subset[:, 0].min() - 0.5, X_subset[:, 0].max() + 0.5
    y_min, y_max = X_subset[:, 1].min() - 0.5, X_subset[:, 1].max() + 0.5
    xx, yy = np.meshgrid(np.linspace(x_min, x_max, 200),
                         np.linspace(y_min, y_max, 200))
    
    # Предсказание модели
    try:
        Z = model.predict_proba(np.c_[xx.ravel(), yy.ravel()])
        if Z.shape[1] == 2:
            Z = Z[:, 1]  # Для бинарной классификации
        else:
            Z = np.argmax(Z, axis=1)  # Для многоклассовой
    except AttributeError:
        Z = model.decision_function(np.c_[xx.ravel(), yy.ravel()])
        if n_classes == 2:
            Z = (Z > 0).astype(int)
        else:
            Z = np.argmax(Z, axis=1)
    Z = Z.reshape(xx.shape)

    # Настройки визуализации
    scatter_params = {'alpha': 0.8, 's': 40} | (scatter_kwargs or {})
    legend_params = {'frameon': False, 'loc': 'best'} | (legend_kwargs or {})
    
    # Создание графика
    fig = plt.figure(figsize=(10, 8), constrained_layout=True)
    gs = gridspec.GridSpec(2, 2, 
                          width_ratios=[4, 1],
                          height_ratios=[1, 4],
                          figure=fig)

    # Основной график
    ax_main = fig.add_subplot(gs[1, 0])
    if len(classes) > 1:
        if Z.ndim == 2:
            contour = ax_main.contourf(xx, yy, Z, alpha=0.3, cmap=cmap)
        else:
            contour = ax_main.contour(xx, yy, Z, alpha=0.6, cmap=cmap)
    else:
        contour = ax_main.contourf(xx, yy, Z, alpha=0.3, cmap=cmap)

    # Точки классов
    for cls in classes:
        mask = y == cls
        label = class_names[cls] if class_names else f'Class {cls}'
        ax_main.scatter(X_subset[mask, 0], X_subset[mask, 1], 
                       label=label, **scatter_params)
        
    # Исправленная установка меток осей
    ax_main.set_xlabel(
        feature_names[0] if feature_names else f'Feature {idx1}'
    )
    ax_main.set_ylabel(
        feature_names[1] if feature_names else f'Feature {idx2}'
    )
    ax_main.legend(**legend_params)

    # Верхний график плотности
    ax_histx = fig.add_subplot(gs[0, 0])
    ax_histx.grid(False)
    for cls in classes:
        subset = X_subset[y == cls, 0]
        kde = gaussian_kde(subset)
        x = np.linspace(x_min, x_max, 100)
        ax_histx.plot(x, kde(x), label=f'Class {cls}')
        ax_histx.fill_between(x, kde(x), alpha=0.3)
    ax_histx.set_axis_off()

    # Правый график плотности
    ax_histy = fig.add_subplot(gs[1, 1])
    ax_histy.grid(False)
    for cls in classes:
        subset = X_subset[y == cls, 1]
        kde = gaussian_kde(subset)
        y_vals = np.linspace(y_min, y_max, 100)
        ax_histy.plot(kde(y_vals), y_vals, label=f'Class {cls}')
        ax_histy.fill_betweenx(y_vals, kde(y_vals), alpha=0.3)
    ax_histy.set_axis_off()

    plt.suptitle('Classification Visualization', y=0.92, fontsize=14)
    plt.show()

PLOT_CLASSIFICATION = [plot_classification_2d]


# интерполяция по Лагранжу
####################################################################################
def lagrange_interpolation(x_values, y_values, x):
    """
    Вычисляет значение интерполяционного многочлена Лагранжа в заданной точке.

    Теоретическая часть:
    Интерполяционный многочлен Лагранжа — это многочлен минимальной степени, принимающий заданные значения в заданном наборе точек [(x₀, y₀), (x₁, y₁), ..., (xₙ₋₁, yₙ₋₁)] [[2]_]. 
    Формула: L(x) = Σₙ₋₁ᵢ₌₀ yᵢ * Πₙ₋₁ⱼ₌₀,ⱼ≠ᵢ (x-xⱼ)/(xᵢ-xⱼ) [[4]_]. 
    Для любого набора из n точек с уникальными xᵢ существует и единственный интерполяционный многочлен степени не выше n-1 [[9]_]. 
    Метод гарантирует прохождение полинома через все заданные точки, но при большом количестве узлов возможен эффект Рунге (сильные осцилляции между узлами) [[6]_].

    Практическая реализация:
    - Проверка корректности входных данных (равные длины x_values и y_values, уникальность элементов в x_values)
    - Вычисление каждого базисного полинома lᵢ(x) по формуле: lᵢ(x) = Πₙ₋₁ⱼ₌₀,ⱼ≠ᵢ (x-xⱼ)/(xᵢ-xⱼ)
    - Суммирование yᵢ·lᵢ(x) для получения финального значения интерполяционного многочлена

    Parameters
    ----------
    x_values : array_like
        Массив узлов интерполяции (координаты x). Все элементы должны быть уникальными.
        Пример: [x₀, x₁, ..., xₙ₋₁]
    y_values : array_like
        Массив значений функции в узлах интерполяции. 
        Пример: [y₀, y₁, ..., yₙ₋₁]
        Длина y_values должна совпадать с длиной x_values.
    x : float
        Точка, в которой вычисляется значение интерполяционного полинома.

    Returns
    -------
    float
        Интерполированное значение в точке x.

    Examples
    --------
    >>> lagrange_interpolation([0, 1, 2], [1, 4, 9], 0.5)
    2.25  # Значение полинома Лагранжа в точке 0.5
    >>> lagrange_interpolation([1, 2, 3], [1, 4, 9], 5)
    25    # Интерполяция за пределами диапазона узлов

    Notes
    -----
    1. Формула Лагранжа реализует полином степени n-1, проходящий через все n заданных точек [[1]_].
    2. При наличии повторяющихся значений в x_values возникнет деление на ноль из-за нарушения условия единственности интерполяции [[9]_].
    3. Эффект Рунге может проявляться при интерполяции большим количеством узлов, особенно для гладких функций [[6]_].

    References
    ----------
    .. [1] "Интерполяционная формула Лагранжа - Википедия", https://en.wikipedia.org/wiki/Lagrange_polynomial
    .. [2] "Интерполяционный многочлен Лагранжа - Prezi", https://example.com/lagrange-prezi
    .. [4] "Интерполяционный многочлен Лагранжа (полином Лагранжа)", https://example.com/lagrange-theory
    .. [6] "Построить интерполяционный многочлен в форме Лагранжа", https://example.com/lagrange-example
    .. [9] "[PDF] Интерполяция многочленом Лагранжа", https://example.com/lagrange-pdf
    """
    n = len(x_values)
    result = 0.0

    # Проверка на совпадение размеров массивов
    if len(y_values) != n:
        raise ValueError("Количество x и y значений должно совпадать")

    # Вычисление суммы по формуле Лагранжа
    for i in range(n):
        term = y_values[i]
        for j in range(n):
            if j != i:
                # Вычисление базисного полинома
                term *= (x - x_values[j]) / (x_values[i] - x_values[j])
        result += term

    return result
####################################################################################
def lagrange_interpolation_func_get(x_values, y_values):
    """
    Построение интерполяционного полинома Лагранжа для заданных точек.

    Теоретическая часть:
    Интерполяционный многочлен Лагранжа представляет собой многочлен степени n-1, проходящий через все n заданных точек [(x₀, y₀), (x₁, y₁), ..., (xₙ₋₁, yₙ₋₁)] [[1]_]. 
    Формула: L(x) = Σₙ₋₁ᵢ₌₀ yᵢ * Πₙ₋₁ⱼ₌₀,ⱼ≠ᵢ (x-xⱼ)/(xᵢ-xⱼ) [[2]_]. 
    Метод гарантирует существование и единственность интерполяционного многочлена при условии, что все узлы интерполяции xᵢ различны [[3]_].

    Практическая реализация:
    - Проверка корректности входных данных (равные длины x_values и y_values, уникальность элементов в x_values)
    - Создание базисных многочленов Лагранжа lᵢ(x) для каждого узла
    - Суммирование yᵢ·lᵢ(x) для получения финального интерполяционного многочлена
    - Возврат функции _lagrange(x), реализующей вычисление значения многочлена в произвольной точке

    Parameters
    ----------
    x_values : array_like
        Список или массив значений координат x, по которым строится интерполяция.
        Все элементы должны быть уникальными. 
        Пример: [x₀, x₁, ..., xₙ₋₁]
    y_values : array_like
        Список или массив значений координат y, соответствующих x_values.
        Пример: [y₀, y₁, ..., yₙ₋₁]
        Длина y_values должна совпадать с длиной x_values.

    Returns
    -------
    callable
        Функция, реализующая интерполяционный полином Лагранжа. 
        Принимает аргумент x (float) и возвращает интерполированное значение (float).

    Examples
    --------
    >>> f = lagrange_interpolation_func_get([0, 1, 2], [1, 4, 9])
    >>> f(0.5)
    2.25  # Значение полинома Лагранжа в точке 0.5
    >>> f(1.5)
    6.25 # Интерполяция между точками (1, 0) и (2, 0)

    Notes
    -----
    1. Метод Лагранжа строит полином степени n-1, проходящий через все n заданных точек [[4]_].
    2. При большом количестве точек возможен эффект Рунге (сильные осцилляции между узлами) [[5]_].
    3. Требование уникальности x_values связано с вычислением базисных полиномов [[6]_].

    References
    ----------
    .. [1] "Интерполяционная формула Лагранжа - Википедия", https://en.wikipedia.org/wiki/Lagrange_polynomial
    .. [2] "Интерполяционный многочлен Лагранжа - Циклопедия", https://example.com/lagrange-cyclopaedia
    .. [3] "Методы интерполяции и аппроксимации - PDF", https://example.com/interpolation-approximation-pdf
    .. [4] Бахвалов Н.С., Жидков Н.П., Кобельков Г.М. Численные методы. М.: Лаборатория Базовых Знаний, 2002.
    .. [5] Kincaid D., Cheney W. Numerical Analysis: Mathematics of Scientific Computing. Brooks/Cole, 2002.
    .. [6] "Интерполяция с помощью формулы Лагранжа - Циклопедия", https://example.com/lagrange-interpolation-cyclopaedia
    """
    if len(x_values) != len(y_values):
        raise ValueError("x_values and y_values must have the same length")
    n = len(x_values)
    
    # Проверка уникальности x_values
    for i in range(n):
        for j in range(i + 1, n):
            if x_values[i] == x_values[j]:
                raise ValueError("x_values must be distinct")
    
    def _basis(i, x):
        """Вычисление базисного многочлена Лагранжа lᵢ(x)"""
        term = 1.0
        xi = x_values[i]
        for j in range(n):
            if j != i:
                xj = x_values[j]
                term *= (x - xj) / (xi - xj)
        return term
    
    def _lagrange(x):
        """Вычисление значения интерполяционного многочлена в точке x"""
        total = 0.0
        for i in range(n):
            total += y_values[i] * _basis(i, x)
        return total
    
    return _lagrange
####################################################################################
def get_delayed_value(history, delay_time, t0):
    """
    Получает значение из истории с линейной интерполяцией для заданного времени, 
    с возможностью экстраполяции на границах. Используется для восстановления 
    значений в системах с задержкой по времени.

    Теоретическое описание:
    Линейная интерполяция вычисляет промежуточные значения между двумя известными точками 
    на основе предположения о линейной зависимости функции на этом участке. Формула:
    f(x) = y0 + (x - x0)*(y1 - y0)/(x1 - x0). При экстраполяции вне диапазона 
    используются крайние значения истории. Это стандартный метод для численного 
    восстановления данных в системах с временной задержкой [[2]][[5]].

    Практическая реализация:
    1. Проверяет корректность входных данных (непустая история, допустимый диапазон времени).
    2. Использует бинарный поиск (bisect_right) для определения соседних точек.
    3. Выполняет линейную интерполяцию между соседними точками или экстраполяцию на границах.

    Parameters
    ----------
    history : list[tuple[float, float]]
        История изменений в формате списка кортежей (время, значение). 
        Должна быть отсортирована по возрастанию времени.
    delay_time : float
        Время, для которого требуется получить значение (обычно t_current - tau).
    t0 : float
        Начальное время системы (используется для обработки случаев delay_time < t0).

    Returns
    -------
    float
        Интерполированное значение для времени delay_time. 
        Возвращает первое значение из истории, если delay_time <= t0 или меньше минимального времени.
        Возвращает последнее значение из истории, если delay_time больше максимального времени.

    Examples
    --------
    >>> history = [(0.0, 1.0), (2.0, 3.0), (4.0, 2.0)]
    >>> get_delayed_value(history, 1.0, 0.0)
    2.0  # Интерполяция между (0,1) и (2,3): y = 1 + (1-0)*(3-1)/(2-0) = 2
    >>> get_delayed_value(history, 5.0, 0.0)
    2.0  # Экстраполяция за пределами истории

    Notes
    -----
    1. История должна быть предварительно отсортирована по времени для корректной работы [[1]].
    2. Метод линейной интерполяции подходит для гладких функций, но может давать погрешность 
       при резких скачках данных [[5]].
    3. Для улучшения точности можно использовать полиномиальные методы интерполяции.

    References
    ----------
    .. [1] "Linear interpolation - Wikipedia" - (https://en.wikipedia.org/wiki/Linear_interpolation)
    .. [2] "Linear Interpolation Formula - BYJU'S" - (https://byjus.com/linear-interpolation-formula/)
    .. [5] "Linear Interpolation Formula: Definition, Formula, Solved Examples" - (https://www.toppr.com/guides/maths-formulas/linear-interpolation-formula/)
    .. [8] "numpy.interp — NumPy v2.2 Manual" - (https://numpy.org/doc/stable/reference/generated/numpy.interp.html)
    """
    from bisect import bisect_right
    if len(history) == 0:
        raise ValueError("История пуста")
    
    if delay_time <= t0:
        return history[0][1]
    
    times = [t for t, _ in history]
    values = [v for _, v in history]
    
    idx = bisect_right(times, delay_time)
    
    if idx == 0:
        return values[0]
    elif idx >= len(times):
        return values[-1]
    else:
        t0_interp = times[idx - 1]
        t1_interp = times[idx]
        v0_interp = values[idx - 1]
        v1_interp = values[idx]
        
        alpha = (delay_time - t0_interp) / (t1_interp - t0_interp)
        return v0_interp * (1 - alpha) + v1_interp * alpha
####################################################################################
INTER = [
    lagrange_interpolation,
    lagrange_interpolation_func_get,
    get_delayed_value]


import requests
from io import BytesIO
from PIL import Image
import IPython.display as display
from ...forall import *

THEORY = []

def list_subdirectories():
    url = "https://api.github.com/repos/Ackrome/matplobblib/contents/pdfs"
    response = requests.get(url)
    if response.status_code == 200:
        contents = response.json()
        return [item['name'] for item in contents if item['type'] == 'dir']
    else:
        print(f"Ошибка при получении подпапок: {response.status_code}")
        return []

def get_png_files_from_subdir(subdir):
    url = f"https://api.github.com/repos/Ackrome/matplobblib/contents/pdfs/{subdir}"
    response = requests.get(url)
    if response.status_code == 200:
        contents = response.json()
        png_files = [item['name'] for item in contents if item['name'].endswith('.png')]
        return [f"https://raw.githubusercontent.com/Ackrome/matplobblib/master/pdfs/{subdir}/{file}" for file in png_files]
    else:
        print(f"Ошибка доступа к {subdir}: {response.status_code}")
        return []

def display_png_files_from_subdir(subdir):
    png_urls = get_png_files_from_subdir(subdir)
    for url in png_urls:
        try:
            response = requests.get(url)
            response.raise_for_status()
            img = Image.open(BytesIO(response.content))
            display.display(img)
        except requests.exceptions.RequestException as e:
            print(f"Ошибка загрузки {url}: {e}")

# Dynamically create functions for each subdirectory
def create_subdir_function(subdir):
    """
    Dynamically creates a function to display PNG files from a given subdirectory.
    The function is named display_png_files_{subdir}.
    """
    global THEORY
    # Define the function dynamically
    def display_function():
        """
        Automatically generated function to display PNG files.
        """
        display_png_files_from_subdir(subdir)
    
    # Set the function name dynamically
    display_function.__name__ = f"display_{subdir}"
    
    # Add a descriptive docstring
    display_function.__doc__ = (
        f"Вывести все страницы из файла с теорией '{subdir.replace('_','-')}'.\n"
        f"Эта функция сгенерирована автоматически из файла '{subdir.replace('_','-')+'.pdf'}' "
        f"из внутрибиблиотечного каталога файлов с теорией."
    )
    
    # Add the function to the global namespace
    globals()[display_function.__name__] = display_function
    THEORY.append(display_function)


# Get subdirectories dynamically
subdirs = list_subdirectories()


# Create functions for each subdirectory dynamically
for subdir in subdirs:
    create_subdir_function(subdir)



text3 ='''
# Функция двойного хэширования
def double_hash_function(self, key):
    return 1 + (hash(key) % (self.size-2))
''' + 'abracadabrabibidi' + '''
# Функция для нахождения первого четного элемента в стеке
def find_first_even(stack):
    current = stack.head
    while current:
        if current.data % 2 == 0:
            return current.data
        current = current.next
    return None
''' + 'abracadabrabibidi' + '''
# Альтернативная функция для нахождения первого четного элемента в стеке
def find_first_even(stack):
    temp_stack = Stack()
    even = None

    while not stack.is_empty():
        item = stack.pop()
        temp_stack.push(item)
        if item % 2 == 0:
            even = item
            break

    while not temp_stack.is_empty():
        stack.push(temp_stack.pop())

    return even
''' + 'abracadabrabibidi' + '''
# Функция для добавления нового элемента в стек после первого нечетного элемента
def add_after_first_odd(stack, item):
    current = stack.head
    while current:
        if current.data % 2 != 0:
            new_node = Node(item)
            new_node.next = current.next
            current.next = new_node
            return
        current = current.next
    stack.push(item)
''' + 'abracadabrabibidi' + '''
# Функция для проверки сбалансированности скобок в математическом выражении
def is_balanced(expr):
    stack = Stack()
    for char in expr:
        if char in "({[":
            stack.push(char)
        elif char in ")}]":
            if stack.is_empty():
                return False
            elif char == ")" and stack.peek() == "(":
                stack.pop()
            elif char == "}" and stack.peek() == "{":
                stack.pop()
            elif char == "]" and stack.peek() == "[":
                stack.pop()
            else:
                return False
    return stack.is_empty()
''' + 'abracadabrabibidi' + '''
# Функция для вычисления математических выражений в обратной польской нотации
def evaluate(expression):
    stack = Stack()
    for token in expression:
        if token.isdigit():
            stack.push(int(token))
        else:
            operand_2 = stack.pop()
            operand_1 = stack.pop()
            if token == '+':
                result = operand_1 + operand_2
            elif token == '-':
                result = operand_1 - operand_2
            elif token == '*':
                result = operand_1 * operand_2
            elif token == '/':
                result = operand_1 / operand_2
            stack.push(result)
    return stack.pop()
''' + 'abracadabrabibidi' + '''
# Функция для нахождения первого нечетного элемента очереди
def find_first_odd(queue):
    current = queue.head
    while current:
        if current.data % 2 != 0:
            return current.data
        current = current.next
    return None
''' + 'abracadabrabibidi' + '''
# Функция для добавления нового элемента в очередь перед первым четным элементом
def add_before_first_even(queue, item):
    new_node = Node(item)
    if not queue.head:
        queue.head = new_node
        queue.tail = new_node
    elif queue.head.data % 2 == 0:
        new_node.next = queue.head
        queue.head = new_node
    else:
        prev_node = queue.head
        current = prev_node.next
        while current:
            if current.data % 2 == 0:
                prev_node.next = new_node
                new_node.next = current
                return
            prev_node = current
            current = current.next
        queue.tail.next = new_node
        queue.tail = new_node
''' + 'abracadabrabibidi' + '''
# Альтернативная функция для добавления нового элемента в очередь перед первым четным элементом
def add_before_first_even(queue, data):
    temp_queue = Queue()
    even_found = False

    while not queue.is_empty():
        item = queue.dequeue()
        if item % 2 == 0 and not even_found:
            temp_queue.enqueue(data)
            even_found = True
        temp_queue.enqueue(item)

    while not temp_queue.is_empty():
        queue.enqueue(temp_queue.dequeue())
''' + 'abracadabrabibidi' + '''
# Функция для удвоения каждого четного элемента двусвязного списка
def double_even_nodes(dllist):
    current_node = dllist.head
    while current_node:
        if current_node.data % 2 == 0:
            new_node = Node(current_node.data)
            new_node.next = current_node.next
            new_node.prev = current_node
            if current_node.next:
                current_node.next.prev = new_node
            current_node.next = new_node
            current_node = new_node.next
        else:
            current_node = current_node.next
''' + 'abracadabrabibidi' + '''
# Функция для удаления всех отрицательных элементов из двусвязного списка
def delete_negative_nodes(dllist):
    current_node = dllist.head
    while current_node:
        if current_node.data < 0:
            if current_node.prev:
                current_node.prev.next = current_node.next
            else:
                dllist.head = current_node.next
            if current_node.next:
                current_node.next.prev = current_node.prev
        current_node = current_node.next
''' + 'abracadabrabibidi' + '''
# Функция, возводящая в квадрат все отрицательные элементы в циклическом двусвязном списке
def square_negative_values(cdllist):
    current_node = cdllist.head
    while current_node:
        if current_node.data < 0:
            current_node.data = current_node.data ** 2
        current_node = current_node.next
        if current_node == cdllist.head:
            break
''' + 'abracadabrabibidi' + '''
# Функция для удаления всех элементов из циклического двусвязного списка, кратных 5
def delete_multiples_of_5(cdllist):
    current_node = cdllist.head
    while current_node:
        if current_node.data % 5 == 0:
            cdllist.delete(current_node.data)
        current_node = current_node.next
        if current_node == cdllist.head:
            break
''' + 'abracadabrabibidi' + '''
# Функция для создания случайного дерева заданной глубины (каждый узел имеет два дочерних узла)
def create_tree(levels):
    tree = Tree()
    nodes = 2**(levels+1) - 1
    values = list(range(1,nodes+1))
    shuffle(values)

    for i in range(nodes):
        value = values[i]
        if i == 0:
            tree.add_node(value)
        else:
            parent_index = (i-1)//2
            parent_value = values[parent_index]
            tree.add_node(value, parent_value)

    return tree
''' + 'abracadabrabibidi' + '''
# Функция для замены каждого числа в дереве на сумму чисел всех его потомков
def replace_with_sum_of_children(tree, node=None):
    if node is None:
        node = tree.root
    if not node.children:
        return node.value
    else:
        sum_of_children = 0
        for child in node.children:
            sum_of_children += replace_with_sum_of_children(tree, child)
        node.value = sum_of_children
        return sum_of_children
''' + 'abracadabrabibidi' + '''
# Функция, удваивающая каждое нечетное число в дереве
def double_odd_values(tree, node=None):
    if node is None:
        node = tree.root
    if node.value % 2 == 1:
        node.value *= 2
    for child in node.children:
        double_odd_values(tree, child)
    return tree
''' + 'abracadabrabibidi' + '''
# Функция для определения листьев дерева
def find_leaves(tree, node=None, leaves=None):
    if leaves is None:
        leaves = []
    if node is None:
        node = tree.root
    if len(node.children) == 0:
        leaves.append(node.value)
    else:
        for child in node.children:
            find_leaves(tree, child, leaves)
    return leaves
''' + 'abracadabrabibidi' + '''
# Функция для нахождения количества узлов в бинарном дереве
def count_nodes(node):
    if node is None:
        return 0
    return 1 + count_nodes(node.left) + count_nodes(node.right)
''' + 'abracadabrabibidi' + '''
# Функция для нахождения всех узлов, которые являются родительскими для заданного узла в бинарном дереве
def find_parents(node, target_node):
    if node is None:
        return []
    if node.left == target_node or node.right == target_node:
        return [node.data]
    left = find_parents(node.left, target_node)
    right = find_parents(node.right, target_node)
    if left:
        return [node.data] + left
    elif right:
        return [node.data] + right
    else:
        return []
''' + 'abracadabrabibidi' + '''
# Функция для нахождения всех узлов, которые имеют значение больше или равно заданному значению в бинарном дереве
def find_nodes(node, value):
    if node is None:
        return []
    result = []
    if node.data >= value:
        result.append(node.data)
    result += find_nodes(node.left, value)
    result += find_nodes(node.right, value)
    return result
''' + 'abracadabrabibidi' + '''
# Функция для нахождения наиболее часто встречающегося значения в хеш-таблице (по полю species)
def most_common_species(hash_table):
    species_count = {}
    for slot in hash_table.table:
        for _, animal in slot:
            if animal.species in species_count:
                species_count[animal.species] += 1
            else:
                species_count[animal.species] = 1
    return max(species_count, key=species_count.get)
'''



import numpy as np
def Kfold_split(n_samples, n_splits=5, shuffle=True, random_state=None):
        """
        Реализация KFold-разбиения.
        
        Параметры:
          n_samples: общее число образцов
          n_splits: число фолдов
          shuffle: перемешивать ли индексы перед разбиением
          random_state: зерно генератора случайных чисел
          
        Возвращает:
          Генератор кортежей (train_indices, test_indices)
        """
        indices = np.arange(n_samples)
        if shuffle:
            rng = np.random.default_rng(random_state)
            rng.shuffle(indices)
        fold_sizes = (n_samples // n_splits) * np.ones(n_splits, dtype=int)
        fold_sizes[:n_samples % n_splits] += 1
        current = 0
        for fold_size in fold_sizes:
            test_indices = indices[current:current + fold_size]
            train_indices = np.concatenate([indices[:current], indices[current + fold_size:]])
            yield train_indices, test_indices
            current += fold_size

FOLDS = [Kfold_split]